module pes.net.http.connection;


private import std.socket;
private import std.string;
private import std.conv;

private import pes.string;
private import pes.net.http.status;
private import pes.net.http.responselistener;
private import pes.net.http.requestgenerator;
private import pes.net.http.header;

class HTTPConnectionException: SocketException
{
	this(char[] msg, int err = 0)
	{
		super(msg, err);
	}
}

class HTTPConnectionTimeoutException: HTTPConnectionException
{
	this()
	{
		super("connection timeout");
	}
}

class HTTPProtocolException: Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}

class HTTPConnection
{
protected:
	TcpSocket				m_socket;
	InternetAddress			m_address;
	HTTPResponseListener	m_listener;
	uint					m_timeout;
	SocketSet 				m_fdset;
	
	int receive(TcpSocket socket, void[] bucket)
	{
		int recv_len, result;
		
		m_fdset.reset();
		m_fdset.add(socket);
		
		if (m_timeout == 0) {
			while ((result = Socket.select(m_fdset, null, null, null)) == -1) {
				// do-nothing
			}
		} else {
			timeval limit;
			limit.seconds 		= m_timeout;
			limit.microseconds	= 0;
			while ((result = Socket.select(m_fdset, null, null, &limit)) == -1) {
				// do-nothing
			}
		}
		switch (result) {
		case 0:		// timeout
			throw new HTTPConnectionTimeoutException();
			
		default:	// OK
			if (!m_fdset.isSet(socket)) {
				close();
				throw new HTTPConnectionException("receive error", std.socket._lasterr());
			}
			break;
		}
		
		recv_len = m_socket.receive(bucket);
		
		return recv_len;
	}
	
	char[] receiveLine(TcpSocket socket)
	{
		int result, count;
		char[] line;
		char[1] c;
		
		count = 0;
		while (true) {
			result = receive(socket, c);
			if (result == c.length) {
				if (c[0] == '\x0a') {
					if (line.length && line[length-1] == '\x0d') {
						line = line[0 .. length-1];
					}
					break;
				}
				line ~= c[0];
			} else {
				break;
			}
		}
		
		return line;
	}
	
	HTTPHeader parseHeader(TcpSocket socket)
	{
		HTTPHeader header = new HTTPHeader;
		char[] line;
		char[][] parts;
		
		line = receiveLine(socket);
		parts = limitedSplit(line, " ", 3);
		if (parts.length != 3 || parts[0].length != 8 /* HTTP/n.n */) {
			throw new HTTPProtocolException("http protocol error1");
		}
		
		header.setVersion		= std.conv.toFloat(parts[0][5 .. length]);
		header.setStatusCode	= toHTTPStatusCode(parts[1]);
		while (true) {
			line = receiveLine(socket);
			if (line.length == 0) {
				break;
			}
			parts = limitedSplit(line, ":", 2);
			if (parts.length != 2) {
				throw new HTTPProtocolException("http protocol error 2");
			}
			header.addField(std.string.strip(parts[0]), std.string.strip(parts[1]));
		}
		
		return header;
	}
	
	HTTPStatusCode toHTTPStatusCode(char[] code)
	{
		HTTPStatusCode status = cast(HTTPStatusCode)std.conv.toUint(code);
		char[] message = getHTTPStatusMessage(status);
		if (message.length == 0) {
			status = HTTPStatusCode.UNDEF;
		}
		
		return status;
	}
	
	void handleResponse(TcpSocket socket, HTTPResponseListener listener, HTTPHeader header)
	{
		bool line_handler_enable	= listener.getLineHandlerEnabled();
		HTTPResponseEvent event 	= new HTTPResponseEvent;
		SocketSet fdset 			= new SocketSet;
		uint bucket_size 			= listener.getBucketSize();
		uint line_pos 				= 0;
		int result, recv_len, end_pos;
		
		void[] bucket;
		char[][] line_buffer;
		
		void handleLine()
		{
			event.reset();
			event.header = header;
			
			// compute length
			uint line_len = 0;
			foreach (e; line_buffer) {
				line_len += e.length;
			}
			
			// copy
			int pos = 0;
			char[] line = new char[line_len];
			
			foreach (e; line_buffer) {
				line[pos .. pos + e.length] = e[0 .. length];
				pos += e.length;
			}
			
			// handle
			event.data = cast(void[])line;
			listener.receivingLine(event);
			if (event.disconnect) {
				close();
			}
			
			// dispose
			delete line;
			line = null;
			foreach (inout e; line_buffer) {
				delete e;
				e = null;
			}
			delete line_buffer;
			line_buffer = null;
		}
		
		if (bucket_size == 0) {
			bucket_size = 1;
		}
		
		bucket 	= cast(void[])new ubyte[bucket_size];
		
		event.reset();
		event.header = header;
		listener.begin(event);	// begin
		
		scope(exit) {
			delete bucket;
			listener.end();	/// end
		}
		
		if (event.disconnect) {
			close();
			return;
		}
		
		while (true) {
			recv_len = receive(m_socket, bucket);
			switch (recv_len) {
			case 0:
				if (line_handler_enable && line_buffer.length != 0) {
					handleLine();
				}
				// don't close
				return;
			case -1:
				close();
				return;
			default:
				break;
			}
			
			// handle
			event.reset();
			event.header 	= header;
			event.data		= bucket[0 .. recv_len];
			listener.receiving(event);
			if (event.disconnect) {
				close();
				if (line_handler_enable && line_buffer.length != 0) {
					handleLine();
				}
				return ;
			}
			
			if (line_handler_enable) {
				line_pos = 0;
				for (int i = 0; i < recv_len; ++i) {
					if ((cast(char[])bucket)[i] == '\x0d'
						|| (cast(char[])bucket)[i] == '\x0a') 
					{
						if (line_buffer.length == 0) {
							event.reset();
							event.header 	= header;
							event.data		= (cast(char[])bucket)[line_pos .. i];
							listener.receivingLine(event);
						} else {
							// compute length
							uint line_len = 0;
							foreach (e; line_buffer) {
								line_len += e.length;
							}
							line_len += i;
							
							// copy
							int pos = 0;
							char[] line = new char[line_len];
							
							foreach (e; line_buffer) {
								line[pos .. pos + e.length] = e[0 .. length];
								pos += e.length;
							}
							line[pos .. pos + i] = (cast(char[])bucket)[0 .. i];
							
							// handle
							event.reset();
							event.header 	= header;
							event.data		= cast(void[])line;
							listener.receivingLine(event);
							
							// dispose
							delete line;
							line = null;
							foreach (inout e; line_buffer) {
								delete e;
								e = null;
							}
							delete line_buffer;
							line_buffer = null;
							
							// disconnect
							if (event.disconnect) {
								close();
								return;
							}
						}
						line_pos = i + 1;
						if (i < recv_len && (cast(char[])bucket)[i + 1] == '\x0a') {
							++i;
							++line_pos;
						}
					}
				}
				if (recv_len != line_pos) {
					line_buffer ~= (cast(char[])bucket)[line_pos .. recv_len].dup;
				}
			}
		}
	}
	
	
public:
	this()
	{
		m_fdset = new SocketSet;
	}
	
	this(InternetAddress address, uint timeout = 0)
	{
		this();
		connect(address, timeout);
	}
	
	void connect(InternetAddress address, uint timeout = 0)
	{
		int result;
		auto SocketSet wrtfd = new SocketSet(1);
		auto SocketSet errfd = new SocketSet(1);
		
		m_socket		= new TcpSocket();
		m_timeout		= timeout;
		
		wrtfd.add(m_socket);
		errfd.add(m_socket);
		m_socket.blocking = false; // non blocking socket
		
		m_socket.connect(address);
		
		if (m_timeout == 0) {
			while ((result = Socket.select(null, wrtfd, errfd, null)) == -1) {
				// do nothing
			}
		} else {
			timeval limit;
			limit.seconds 		= m_timeout;
			limit.microseconds	= 0;
			
			while ((result = Socket.select(null, wrtfd, errfd, &limit)) == -1) {
				// do nothing
			}
		}
		
		switch (result) {
		case 0:		// timeout
			throw new HTTPConnectionTimeoutException();
			
		default:	// OK
			if (errfd.isSet(m_socket)) {
				// error
				throw new HTTPConnectionException("Unable to connect socket", result);
			}
			break; // success
		}
	}
	
	void io(HTTPRequestGenerator generator)
	{
		void[] request = generator.generateRequest();
		
		m_socket.send(request);
		delete request;
		request = null;
		
		if (m_listener is null) {
			int result;
			void[] bucket = new void[0x10000];
			while ((result = receive(m_socket, bucket)) > 0) {
				// do-nothing
			}
			delete bucket;
			close();
			return;
		}
		
		handleResponse(m_socket, m_listener, parseHeader(m_socket));
	}
	
	void close()
	{
		if (m_socket !is null) {
			m_socket.close();
			m_socket = null;
		}
	}
	
	void setListener(HTTPResponseListener listener)
	{
		m_listener = listener;
	}
	
	void removeListener()
	{
		m_listener = null;
	}
}

debug(pes_net_http_io)
{
	import std.cstream;
	import std.stdio;
	import pes.net.http.singlegetrequest;
	class Listener: HTTPResponseListener
	{
		this()
		{
			setLineHandlerEnabled = true;
		}
		void receiving(HTTPResponseEvent event)
		{
			writef("receiving: %s\n", event.text.length);
		}
		void receivingLine(HTTPResponseEvent event) 
		{
			writef("receivingLine: %s\n", event.text);
		}
	}
	
	int main()
	{
		SingleGetRequest request = new SingleGetRequest("/server.cgi/get/thread_E4BB8AE5BE8CE382B9E383AC/0-");
		HTTPConnection conn = new HTTPConnection;
		
		
		request.getHeader.addField("Host", "localhost");
		request.getHeader.addField("X-Host", "localhost");
		writef("send--\n%s\n", request.toString);
		
		try {
			conn.setListener(new Listener);
			conn.connect(new InternetAddress("127.0.0.1", 8000));
			conn.io(request);
			conn.close();
		} catch (Exception e) {
			writef("--Exception: %s\n", e.toString);
		}
		din.readLine();
		return 0;
	}
}

