module pes.net.shingetsu.serverio;

private import std.string;
private import std.conv;
private import std.socket;
private import std.c.time;

private import pes.string;
private import pes.conv;
private import pes.net.http.all;
private import pes.net.shingetsu.get;
private import pes.net.shingetsu.util;

class ThreadDataException: Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}

class ThreadDataEvent
{
protected:
	const char[]	DATA_DELIMITER = "<>";
	const char[]	ELEMENT_DELIMITER = ":";
	
public:
	time_t			time;
	char[]			md5;
	char[][char[]]	elements;
	Object			cData;
	
	this()
	{
	}
	
	void parse(char[] line)
	{
		char[][] parts = std.string.split(line, DATA_DELIMITER);
		if (parts.length < 2) {
			throw new ThreadDataException("thread data error 1");
		}
		time 	= cast(time_t)std.conv.toUint(parts[0]);
		
		if (parts[1].length != 32) {
			throw new ThreadDataException("thread data error2");
		}
		md5 = parts[1].dup;
		
		if (parts.length > 2) {
			parts = parts[2 .. length]; // splice
			foreach (element; parts) {
				char[][] elm_parts = limitedSplit(element, ELEMENT_DELIMITER, 2);
				if (elm_parts.length != 2) {
					throw new ThreadDataException("thread data error3");
				}
				elements[elm_parts[0].dup] = elm_parts[1].dup; // duplicate
			}
		}
	}
	
	void dispose()
	{
		clear();
	}
	
	void clear()
	{
		if (md5.length == 0) {
			delete md5;
		}
		md5 = null;
		
		foreach (char[] key; elements.keys) {
			elements.remove(key);
		}
		elements = null;
		time = 0;
	}
}

class ServerIO: ShinGETsuIO
{
protected:
	
	class RangeGetRequest: SingleGetRequest
	{
		this(char[] thread_name, time_t begin = 0) 
		{
			super(std.string.format("/server.cgi/get/%s/%s-",
				encodeThreadName(thread_name), cast(uint)begin));
		}
	}
	
	class HeadRequest: SingleGetRequest
	{
		this(char[] thread_name, time_t begin = 0) {
			super(std.string.format("/server.cgi/head/%s/%s-",
				encodeThreadName(thread_name), cast(uint)begin));
		}
	}
	
	class HashGetRequest: SingleGetRequest
	{
		this(char[] thread_name, time_t time, char[] md5)
		{
			super(std.string.format("/server.cgi/get/%s/%s/%s",
				encodeThreadName(thread_name), cast(uint)time, md5));
		}
	}
	
	class HaveRequest: SingleGetRequest
	{
		this(char[] thread_name)
		{
			super(std.string.format("/server.cgi/have/%s", encodeThreadName(thread_name)));
		}
	}
	
	class HaveThreadResponseListener: HTTPResponseListener
	{
	protected:
		bool	m_has_thread 	= false;
		bool	m_success		= false;
		
	public:
		this()
		{
			setLineHandlerEnabled(true);
		}
		
		override void receivingLine(HTTPResponseEvent event)
		{
			char[] text = event.text;
			if (std.string.icmp(event.text, "YES") == 0) {
				m_has_thread 	= true;
				m_success		= true;
			} else if (std.string.icmp(event.text, "NO") == 0) {
				m_has_thread 	= false;
				m_success		= true;
			} else {
				m_success		= false;
			}
		}
		
		override void receiving(HTTPResponseEvent event)
		{
			char[] text = event.text;
			
			// do nothing
		}
		
		bool isSuccessful()
		{
			return m_success;
		}
		
		bool hasThread()
		{
			return m_has_thread;
		}
	}
	
	
	class ThreadDataResponseListener: HTTPResponseListener
	{
	protected:
		ThreadDataEvent		 				m_event;
		void delegate(ThreadDataEvent)		m_handler;
		Object								m_argv;
	public:
		
		this(void delegate(ThreadDataEvent) handler, Object data)
		{
			m_handler 	= handler;
			m_argv		= data;
			
			setLineHandlerEnabled(true);
			
			m_event = new ThreadDataEvent;
			m_event.clear();
		}
		
		override void receivingLine(HTTPResponseEvent event)
		{
			m_event.parse(event.text);
			m_event.cData = m_argv;
			
			m_handler(m_event);
			
			m_event.clear();
		}
		
		override void receiving(HTTPResponseEvent event)
		{
			// do nothing
		}
	}
	
public:
	this(InternetAddress address)
	{
		super(address);
	}
	
	void getThreadData(Object argv, 
						void delegate(ThreadDataEvent) handler,
						char[] thread_name, time_t begin = 0)
	{
		RangeGetRequest 			generator 	= new RangeGetRequest(thread_name, begin);
		ThreadDataResponseListener 	listener	= new ThreadDataResponseListener(handler, argv);
		
		scope(exit) {
			removeListener();
			delete listener;
			delete generator;
		}
		setListener(listener);
		
		get(generator);
	}
	
	void getThreadData(Object argv, 
						void delegate(ThreadDataEvent) handler,
						char[] thread_name, time_t time, char[] md5)
	{
		HashGetRequest 				generator 	= new HashGetRequest(thread_name, time, md5);
		ThreadDataResponseListener 	listener	= new ThreadDataResponseListener(handler, argv);
		
		scope(exit) {
			removeListener();
			delete listener;
			delete generator;
		}
		setListener(listener);
		
		get(generator);
	}
	
	bool haveThread(char[] thread_name)
	{
		HaveRequest 				generator 	= new HaveRequest(thread_name);
		HaveThreadResponseListener 	listener	= new HaveThreadResponseListener;
		
		scope(exit) {
			removeListener();
			delete listener;
			delete generator;
		}
		setListener(listener);
		
		get(generator);
		
		if (!listener.isSuccessful) {
			new Exception("have response error");
		}
		
		return listener.hasThread();
	}
	
	void getThreadHeader(Object argv,
							void delegate(ThreadDataEvent) handler,
							char[] thread_name, time_t begin = 0)
	{
		HeadRequest 				generator 	= new HeadRequest(thread_name, begin);
		ThreadDataResponseListener 	listener	= new ThreadDataResponseListener(handler, argv);
		
		scope(exit) {
			removeListener();
			delete listener;
			delete generator;
		}
		setListener(listener);
		
		get(generator);
	}
}

