/**
 * キャッシュDB
 *
 */
module threadcache;


private import std.c.time;
private import std.file;
private import std.path;
private import pes.path;

private import pes.conv;
private import pes.checksum;
private import pes.data.sqlite3;
private import pes.net.shingetsu.all;
private import config;
private import schema;

private const char[] CREATE_CACHE_TABLE =
"CREATE TABLE THREAD_CACHE (
	TITLE		 		TEXT,
	MTIME 				INT,
	MD5 				TEXT,
	MAIL				TEXT,
	NAME 				TEXT,
	MESSAGE 			TEXT,
	TRIP 				TEXT,
	REMOVE_ID			TEXT,
	REMOVE_STAMP		TEXT,
	PUBKEY 				TEXT,
	SIGN 				TEXT,
	ATTACH 				TEXT,
	EXT 				TEXT,
	FLAGS				INT
);";

private const char[] CREATE_INDEX_TABLE =
"CREATE TABLE THREAD_INDEX (
	TITLE				TEXT,
	PTIME				INT,
	PHASH				TEXT,
	ATIME				INT,
	COUNT				INT,
	FLAG				INT
);";

private const char[] CREATE_TINDEX =
"CREATE INDEX PINDEX ON THREAD_CACHE(TITLE);";
private const char[] CREATE_PINDEX =
"CREATE INDEX PINDEX2 ON THREAD_CACHE(TITLE, MTIME);";
private const char[] CREATE_FINDEX = 
"CREATE INDEX FINDEX ON THREAD_CACHE(TITLE, FLAGS);";
private const char[] CREATE_LINDEX =
"CREATE UNIQUE INDEX LINDEX ON THREAD_INDEX(TITLE)";
private const char[] CREATE_UINDEX =
"CREATE UNIQUE INDEX UINDEX ON THREAD_CACHE(TITLE,MTIME,MD5);";

private const char[] EXISTS_THREAD_INDEX =
"SELECT COUNT(*) FROM THREAD_INDEX WHERE TITLE=?;";	// 0 OR EXISTS

private const char[] INSERT_THREAD_INDEX =
"INSERT INTO THREAD_INDEX (TITLE,PTIME,PHASH,ATIME,COUNT,FLAG)
	VALUES(
		?, 
		(SELECT MTIME FROM THREAD_CACHE WHERE TITLE=? ORDER BY MTIME DESC LIMIT 1),
		(SELECT MD5 FROM THREAD_CACHE WHERE TITLE=? ORDER BY MTIME DESC LIMIT 1),
		?,
		(SELECT COUNT(*) FROM THREAD_CACHE WHERE TITLE=?),
		1
	);";

private const char[] UPDATE_THREAD_INDEX =
"UPDATE THREAD_INDEX 
	SET 
		PTIME = (SELECT MTIME FROM THREAD_CACHE WHERE TITLE=? ORDER BY MTIME DESC LIMIT 1),
		PHASH = (SELECT MD5 FROM THREAD_CACHE WHERE TITLE=? ORDER BY MTIME DESC LIMIT 1),
		ATIME = ?,
		COUNT = (SELECT COUNT(*) FROM THREAD_CACHE WHERE TITLE=?),
		FLAG  = FLAG|1
	WHERE TITLE=?;"; // AND, ACTIVATE FLAG(0x00000001)

private const char[] REMOVE_FLAG_INDEX =
"UPDATE THREAD_INDEX SET FLAG=FLAG&~1 WHERE TITLE=?";


private const char[] INSERT_CACHE =
"INSERT INTO THREAD_CACHE (TITLE,MTIME,MD5,MAIL,NAME,MESSAGE,TRIP,REMOVE_ID,REMOVE_STAMP,PUBKEY,SIGN,ATTACH,EXT,FLAGS)
	VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?);";
	
private const char[] GET_INDEX =
"SELECT * FROM THREAD_INDEX WHERE TITLE=?;";

private const char[] GET_INDEXES =
"SELECT * FROM THREAD_INDEX ORDER BY PTIME DESC;";

private const char[] GET_RECENT_INDEXES =
"SELECT * FROM THREAD_INDEX WHERE PTIME >= ? ORDER BY PTIME DESC;";

private const char[] GET_RECENT_SUMMARIES =
"SELECT MTIME,MD5 FROM THREAD_CACHE WHERE TITLE=? AND MTIME >= ?;";

private const char[] GET_MESSAGE =
"SELECT * FROM THREAD_CACHE WHERE TITLE=? AND MD5 LIKE(? || '%');";

private const char[] GET_MESSAGES =
"SELECT * FROM THREAD_CACHE WHERE TITLE=? ORDER BY MTIME;";

private const char[] GET_MESSAGE_COUNT =
"SELECT count(*) FROM THREAD_CACHE WHERE TITLE=?;";

/** スレッドインデックステーブル */
class ThreadIndex
{
	char[]	title;					// タイトル
	time_t	lastPostTime;			// 最終投稿日時
	char[]	lastPostHash;			// 最終投稿HASH
	time_t	lastAccessTime;			// 最終アクセス日時
	uint	count;					// 投稿数
	uint	flag;					// 1ビット目=登録済み:1/削除済み:0
	
	char[] id()
	{
		if (lastPostHash.length >= 8) {
			return lastPostHash[0 .. 8];
		}
		
		return null;
	}
	
	void dispose()
	{
		if (title.length != 0) {
			delete title;
			title = null;
		}
		if (lastPostHash.length != 0) {
			delete lastPostHash;
			lastPostHash = null;
		}
	}
}

/** メッセージサマリ */
class ThreadMessageSummary
{
	time_t	time;			// 投稿日時
	char[]	md5;			// メッセージのMD5値
	
	void dispose()
	{
		if (md5.length != 0) {
			delete md5;
			md5 = null;
		}
	}
}

/** メッセージ */
class ThreadMessage
{
	char[]	title;			// タイトル
	time_t	time;			// 投稿日時
	char[]	md5;			// メッセージのMD5値
	char[]	message;		// メッセージ
	char[] 	mail;			// メール
	char[]	name;			// 投稿者名
	char[]	removeId;		// 削除ID
	char[]	removeStamp;	// 削除タイムスタンプ
	char[]  trip;			// トリップ
	char[]	pubkey;			// 公開鍵
	char[]	sign;			// 署名
	char[]	attach;			// 添付ファイルMD5SUM (ない場合は空白)
	char[]	ext;			// 添付ファイルの拡張子
	uint	flag;			// 1bit目=読んだ
	
	char[] getAttachPath(ubyte[] attachment)
	{
		return std.path.join(absPath(getConfig.attachmentsDirectory),
			std.string.format("%s.%s", binaryToHexString(md5Sum(attachment)), ext));
	}
	
	void setAttachName(ubyte[] attachment)
	{
		attach = binaryToHexString(md5Sum(attachment));
	}
	
	char[] getAttachPath()
	{
		if (attach is null) return null;
		return std.path.join(absPath(getConfig.attachmentsDirectory),
								std.string.format("%s.%s", attach, ext));
	}
	
	char[] toXMLItem(uint no)
	{
		char[] escaping;
		
		escaping = std.string.replace(message, "&", "&amp;");
		escaping = std.string.replace(escaping, "<br>", "</LINE><LINE>");
		escaping = std.string.replace(escaping, "\t", "    ");
		escaping = std.string.format("<LINE>%s</LINE>", escaping);
		
		return std.string.format(
`		<ITEM>
			<ID>%s</ID>
			<NO>%s</NO>
			<HASH>%s</HASH>
			<TIMESTAMP>%s</TIMESTAMP>
			<NAME>%s</NAME>
			<MAIL>%s</MAIL>
			<TRIP>%s</TRIP>
			<REMOVE_ID>%s</REMOVE_ID>
			<REMOVE_STAMP>%s</REMOVE_STAMP>
			<MESSAGE>%s</MESSAGE>
			<ATTACH_PATH>%s</ATTACH_PATH>
			<ATTACH_HASH>%s</ATTACH_HASH>
			<ATTACH_EXT>%s</ATTACH_EXT>
			<FLAGS>%s</FLAGS>
		</ITEM>
`
			, md5[0 .. 8], no, md5, time,  name, mail, trip
			, removeId, removeStamp
			, escaping
			, getAttachPath(), attach, ext, flag
		);
	}
	
	static char[] toXML(Schema schema, char[] theme_dir,
						 ThreadMessage[] messages, uint begin_no, char[] recent_id)
	{
		uint	no			= begin_no;
		uint 	count		= messages.length;
		char[]	xslt;
		
		xslt	= Configuration.xslTemplate(theme_dir, schema.method);
		if (xslt !is null) {
			xslt = std.string.format(`<?xml-stylesheet type="text/xsl" href="%s" ?>`, xslt);
		}
		
		char[] 	header		= std.string.format(
`<?xml version="1.0"?>
%s
<SHINGETSU>
	<INFO>
		<TITLE>%s</TITLE>
		<FILE>%s</FILE>
		<TARGET>%s</TARGET>
		<GATEWAY>%s</GATEWAY>
		<METHOD>%s</METHOD>
		<BEGINNO>%s</BEGINNO>
		<ENDNO>%s</ENDNO>
		<COUNT>%s</COUNT>
		<RECENT_ID>%s</RECENT_ID>
		<THEME>%s</THEME>
	</INFO>
	<ITEMS>
`
			, xslt
			, schema.threadName, encodeThreadName(schema.threadName), schema.hash
			, getConfig.gatewayString, schema.method
			, begin_no, begin_no + count - (count ? 1:0), count, recent_id, theme_dir
		);
		char[] footer =
`	</ITEMS>
</SHINGETSU>
`
		;
		char[] items, item, page;
		foreach (e; messages) {
			item	= e.toXMLItem(no++);
			items ~= item;
			item = null;
//			delete item;
		}
		page = std.string.format("%s%s%s", header, items, footer);
//		delete header;
//		delete items;
//		delete footer;
		
		return page;
	}
	
	void dispose()
	{
		if (title.length != 0) {
//			delete title;
			title = null;
		}
		if (md5.length != 0) {
//			delete md5;
			md5 = null;
		}
		if (message.length != 0) {
//			delete message;
			message = null;
		}
		if (name.length != 0) {
//			delete name;
			name = null;
		}
		if (mail.length != 0) {
//			delete mail;
			mail = null;
		}
		if (pubkey.length != 0) {
//			delete pubkey;
			pubkey = null;
		}
		if (trip.length != 0) {
//			delete trip;
			trip = null;
		}
		if (sign.length != 0) {
//			delete sign;
			sign = null;
		}
		if (ext.length != 0) {
//			delete ext;
			ext = null;
		}
	}
}

/** スレッドキャッシュテーブル */
class ThreadCache
{
private:
	static ThreadCache g_thread_cache;

protected:
	Object	m_lock;
	SQLite	m_db;
	
	ThreadMessage fetchMessage(SQLiteStatement stmt)
	{
		ThreadMessage message = new ThreadMessage;
		int pos = 0;
		
		message.title		= stmt.fetch!(char[])(pos++).dup;
		message.time		= stmt.fetch!(time_t)(pos++);
		message.md5			= stmt.fetch!(char[])(pos++).dup;
		message.mail		= stmt.fetch!(char[])(pos++).dup;
		message.name		= stmt.fetch!(char[])(pos++).dup;
		message.message		= stmt.fetch!(char[])(pos++).dup;
		message.trip		= stmt.fetch!(char[])(pos++).dup;
		message.removeId	= stmt.fetch!(char[])(pos++).dup;
		message.removeStamp	= stmt.fetch!(char[])(pos++).dup;
//		message.pubkey		= stmt.fetch!(char[])(pos++).dup;
		++pos;
//		message.sign		= stmt.fetch!(char[])(pos++).dup;
		++pos;
		
		message.attach		= stmt.fetch!(char[])(pos++).dup;
		message.ext			= stmt.fetch!(char[])(pos++).dup;
		message.flag		= stmt.fetch!(uint)(pos++);
		
		return message;
	}
	
	ThreadMessageSummary fetchMessageSummary(SQLiteStatement stmt)
	{
		ThreadMessageSummary summary = new ThreadMessageSummary;
		int pos = 0;
		
		summary.time	= stmt.fetch!(time_t)(pos++);
		summary.md5		= stmt.fetch!(char[])(pos++).dup;
		
		return summary;
	}
	
	ThreadIndex fetchIndex(SQLiteStatement stmt)
	{
		ThreadIndex index = new ThreadIndex;
		int pos = 0;
		
		index.title				= stmt.fetch!(char[])(pos++).dup;
		index.lastPostTime		= stmt.fetch!(time_t)(pos++);
		index.lastPostHash		= stmt.fetch!(char[])(pos++).dup;
		index.lastAccessTime	= stmt.fetch!(time_t)(pos++);
		index.count				= stmt.fetch!(uint)(pos++);
		index.flag				= stmt.fetch!(uint)(pos++);
		
		return index;
	}
	
	this()
	{
		m_lock = new Object;
		m_db = new SQLite(getConfig().cacheFile);
		
		synchronized(lockObject) {
			try {
				m_db.exec(CREATE_CACHE_TABLE);
				m_db.exec(CREATE_INDEX_TABLE);
				
				m_db.exec(CREATE_TINDEX);
				m_db.exec(CREATE_PINDEX);
				m_db.exec(CREATE_FINDEX);
				m_db.exec(CREATE_LINDEX);
				m_db.exec(CREATE_UINDEX);
			} catch (Exception) {
				// do-nothing
			}
		}
	}
	
	static this()
	{
		g_thread_cache = new ThreadCache();
	}
	
	time_t getTimeStamp()
	{
		return time(null) - (3600 * 16);
	}
public:

	
	static ThreadCache getInstance()
	{
		return g_thread_cache;
	}
	
	ThreadIndex getIndex(char[] title)
	{
		synchronized(lockObject) 
		{
			SQLiteStatement stmt;
			
			m_db.begin();
			stmt = m_db.prepare(GET_INDEX);
			scope(exit)
			{
				if (stmt !is null) {
					stmt.finalize();
				}
				m_db.end();
			}
			
			stmt.bind(1, title);
			stmt.step();
			
			if (stmt.hasRow) {
				return fetchIndex(stmt);
			}
			
			return null;
		}
	}
	
	ThreadIndex[] getIndexes()
	{
		ThreadIndex[] indexes;
		
		synchronized(lockObject) 
		{
			SQLiteStatement stmt;
			
			m_db.begin();
			stmt = m_db.prepare(GET_INDEXES);
			scope(exit)
			{
				if (stmt !is null) {
					stmt.finalize();
				}
				m_db.end();
			}
			
			for (stmt.step(); stmt.hasRow();stmt.step()) {
				indexes ~= fetchIndex(stmt);
			}
		}
		
		return indexes;
	}
	
	ThreadIndex[] getRecentIndexes(time_t base = 0)
	{
		ThreadIndex[] indexes;
		
		synchronized(lockObject) 
		{
			SQLiteStatement stmt;
			
			m_db.begin();
			stmt = m_db.prepare(GET_RECENT_INDEXES);
			scope(exit)
			{
				if (stmt !is null) {
					stmt.finalize();
				}
				m_db.end();
			}
			
			stmt.bind(1, base);
			for (stmt.step(); stmt.hasRow();stmt.step()) {
				indexes ~= fetchIndex(stmt);
			}
		}
		
		return indexes;
	}
	
	ThreadMessageSummary[] getRecentMessageSummaries(char[] title, time_t base = 0)
	{
		ThreadMessageSummary[] summaries;
		
		synchronized(lockObject) 
		{
			SQLiteStatement stmt;
			
			m_db.begin();
			stmt = m_db.prepare(GET_RECENT_SUMMARIES);
			scope(exit)
			{
				if (stmt !is null) {
					stmt.finalize();
				}
				m_db.end();
			}
			
			stmt.bind(1, title, base);
			for (stmt.step(); stmt.hasRow(); stmt.step()) {
				summaries ~= fetchMessageSummary(stmt);
			}
		}
		
		return summaries;
	}
	
	uint getMessageCount(char[] title)
	{
		uint count;
		
		synchronized(lockObject) 
		{
			SQLiteStatement stmt;
			
			m_db.begin();
			stmt = m_db.prepare(GET_MESSAGE_COUNT);
			scope(exit)
			{
				if (stmt !is null) {
					stmt.finalize();
				}
				m_db.end();
			}
			
			stmt.bind(1, title);
			stmt.step();
			if (stmt.hasRow()) {
				count = stmt.fetch!(uint)(0);
			}
		}
		
		return count;
	}
	
	ThreadMessage getMessage(char[] title, char[] hash_front)
	{
		ThreadMessage message;
		
		synchronized(lockObject) 
		{
			SQLiteStatement stmt;
			
			m_db.begin();
			stmt = m_db.prepare(GET_MESSAGE);
			scope(exit)
			{
				if (stmt !is null) {
					stmt.finalize();
				}
				m_db.end();
			}
			
			stmt.bind(1, title, hash_front);
			
			uint i = 0;
			stmt.step();
			
			if (stmt.hasRow()) {
				message = fetchMessage(stmt);
			}
		}
		
		return message;
	}
	
	ThreadMessage[] getMessages(char[] title, uint begin, uint count = uint.max)
	{
		if (count == 0) {
			return null;
		}
		ThreadMessage[] messages;
		
		synchronized(lockObject) 
		{
			SQLiteStatement stmt;
			
			m_db.begin();
			stmt = m_db.prepare(GET_MESSAGES);
			scope(exit)
			{
				if (stmt !is null) {
					stmt.finalize();
				}
				m_db.end();
			}
			
			stmt.bind(1, title);
			
			uint i = 0;
			for (stmt.step(); stmt.hasRow(); stmt.step()) {
				if (begin <= i) {
					if (count-- == 0) {
						break;
					}
					messages ~= fetchMessage(stmt);
				}
				++i;
			}
		}
		
		return messages;
	}
	
	
	void updateIndex(char[] title)
	{
		synchronized(lockObject) 
		{
			SQLiteStatement exists_stmt;
			SQLiteStatement update_stmt;
			
			m_db.begin();
			exists_stmt = m_db.prepare(EXISTS_THREAD_INDEX);
			
			scope(exit) { // SCOPE EXIT
				if (exists_stmt !is null) {
					exists_stmt.finalize();
				}
				if (update_stmt !is null) {
					update_stmt.finalize();
				}
				m_db.commit();
			}
			
			exists_stmt.bind(1, title);
			exists_stmt.step();
			
			if (exists_stmt.hasRow() && exists_stmt.fetch!(uint)(0)) {
				// already exists
				exists_stmt.finalize();
				exists_stmt = null;
				
				update_stmt = m_db.prepare(UPDATE_THREAD_INDEX);
				update_stmt.bind(1, title, title, getTimeStamp(), title, title);
			} else {
				// does not exists
				exists_stmt.finalize();
				exists_stmt = null;
				
				update_stmt = m_db.prepare(INSERT_THREAD_INDEX);
				update_stmt.bind(1, title, title, title, getTimeStamp(), title);
			}
			update_stmt.step();
		}
	}
	
	void removeIndex(char[] title)
	{
		synchronized(lockObject) 
		{
			SQLiteStatement stmt;
			
			m_db.begin();
			stmt = m_db.prepare(REMOVE_FLAG_INDEX);
			scope(exit)
			{
				if (stmt !is null) {
					stmt.finalize();
				}
				m_db.commit();
			}
			
			stmt.bind(1, title);
			stmt.step();
		}
	}
	
	void insertMessage(ThreadMessage message)
	{
		ThreadMessage[] messages;
		
		messages ~= message;
		insertMessages(messages);
	}
	
	void insertMessages(ThreadMessage[] messages)
	{
		synchronized(lockObject)
		{
			// insert messages into cache
			{// scope 
				SQLiteStatement stmt;
				m_db.begin();
				stmt = m_db.prepare(INSERT_CACHE);
				
				scope(exit) {
					if (stmt !is null) {
						stmt.finalize();
					}
					m_db.commit();
				}
				
				foreach (e; messages) {
					try {
						stmt.bind(1,
								e.title, e.time, e.md5,
								e.mail, e.name, e.message, e.trip,
								e.removeId, e.removeStamp,
								e.pubkey, e.sign, e.attach, e.ext, e.flag
						);
						stmt.step(); // insert
					} catch (SQLiteException) {
						// unique index error
					} finally {
						stmt.reset();
					}
				}
			}
			// update indexes
			int[char[]] update_table;
			foreach (e; messages) {
				if ((e.title in update_table) == null) {
					update_table[e.title] = 1;
					updateIndex(e.title);
				}
			}
		}
	}
	
	Object lockObject()
	{
		return m_lock;
	}
}










